// Mind's Eye by Mark Henry

// The following include stdlib.h must be before the glut includes
#include <stdlib.h>

#include <GL/glew.h>
#include <GL/glut.h>

#include <iostream>
#include <string>

#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>

#include "GLSL_helper.h"
#include "MStackHelp.h"
#include "GeometryCreator.h"

#include "bass.h"
#include "bass_fx.h"

using namespace std;
using namespace glm;

// Parameters
unsigned int time = 0;
unsigned int timeLastBeat = 1000;
unsigned int WindowWidth = 800, WindowHeight = 600;

// Variable Handles
GLuint aPosition;
GLuint aNormal;
GLuint uModelMatrix;
GLuint uNormalMatrix;
GLuint uViewMatrix;
GLuint uProjMatrix;
GLuint uColor;

// Shader Handle
GLuint ShadeProg;

RenderingHelper ModelTrans;

// BASS
HWND win=NULL;
DWORD chan;
DWORD hBPM;
float orgBPM;

void SetProjectionMatrix()
{
    glm::mat4 Projection = glm::perspective(80.0f, ((float) WindowWidth)/ ((float)WindowHeight), 0.1f, 100.f);
    safe_glUniformMatrix4fv(uProjMatrix, glm::value_ptr(Projection));
}

void SetView()
{
   glm::mat4 View = glm::lookAt(vec3(0,0,3), vec3(0,0,0), vec3(0,1,0));
   safe_glUniformMatrix4fv(uViewMatrix, glm::value_ptr(View));
}

void SetModel()
{
    safe_glUniformMatrix4fv(uModelMatrix, glm::value_ptr(ModelTrans.modelViewMatrix));
    safe_glUniformMatrix4fv(uNormalMatrix, glm::value_ptr(glm::transpose(glm::inverse(ModelTrans.modelViewMatrix))));
}

bool InstallShader(std::string const & vShaderName, std::string const & fShaderName)
{
    GLuint VS; // handles to shader object
    GLuint FS; // handles to frag shader object
    GLint vCompiled, fCompiled, linked; // status of shader

    VS = glCreateShader(GL_VERTEX_SHADER);
    FS = glCreateShader(GL_FRAGMENT_SHADER);

    // load the source
    char const * vSource = textFileRead(vShaderName);
    char const * fSource = textFileRead(fShaderName);
    glShaderSource(VS, 1, & vSource, NULL);
    glShaderSource(FS, 1, & fSource, NULL);

    // compile shader and print log
    glCompileShader(VS);
    printOpenGLError();
    glGetShaderiv(VS, GL_COMPILE_STATUS, & vCompiled);
    printShaderInfoLog(VS);

    // compile shader and print log
    glCompileShader(FS);
    printOpenGLError();
    glGetShaderiv(FS, GL_COMPILE_STATUS, & fCompiled);
    printShaderInfoLog(FS);

    if (! vCompiled || ! fCompiled)
    {
        std::cerr << "Error compiling either shader " << vShaderName << " or " << fShaderName << std::endl;
        return false;
    }

    // create a program object and attach the compiled shader
    ShadeProg = glCreateProgram();
    glAttachShader(ShadeProg, VS);
    glAttachShader(ShadeProg, FS);

    glLinkProgram(ShadeProg);

    // check shader status requires helper functions
    printOpenGLError();
    glGetProgramiv(ShadeProg, GL_LINK_STATUS, &linked);
    printProgramInfoLog(ShadeProg);

    glUseProgram(ShadeProg);

    // get handles to attribute data
    aPosition   = safe_glGetAttribLocation(ShadeProg, "aPosition");
    aNormal     = safe_glGetAttribLocation(ShadeProg, "aNormal");
    
    uColor          = safe_glGetUniformLocation(ShadeProg, "uColor");
    uProjMatrix     = safe_glGetUniformLocation(ShadeProg, "uProjMatrix");
    uViewMatrix     = safe_glGetUniformLocation(ShadeProg, "uViewMatrix");
    uModelMatrix    = safe_glGetUniformLocation(ShadeProg, "uModelMatrix");
    uNormalMatrix   = safe_glGetUniformLocation(ShadeProg, "uNormalMatrix");

    std::cout << "Sucessfully installed shader " << ShadeProg << std::endl;
    return true;
}

void Initialize()
{
    glClearColor(0, 0, 0, 1);

    glClearDepth(1.0f);
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_DEPTH_TEST);

    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);

    ModelTrans.useModelViewMatrix();
    ModelTrans.loadIdentity();
}

// Calculates rgb from hue, sat, val
// Adapted from http://en.wikipedia.org/wiki/HSL_and_HSV#Converting_to_RGB
void getRGB(int hue, int sat, int val, int *r, int *g, int *b)
{
	// hue: 0-355, sat: 0-255, val (lightness): 0-255
   *r = *g = *b = 0;

   float hdash = hue / 60.0;
   int chroma = (int)((val / 256.0) * (sat / 256.0) * 256);
   int x = chroma * (1 - abs(fmod(hdash, 2) - 1));
   switch (hue / 60)
   {
   case 0:
      *r = chroma;
      *g = x;
      *b = 0;
   case 1:
      *r = x;
      *g = chroma;
      *b = 0;
   case 2:
      *r = 0;
      *g = chroma;
      *b = x;
   case 3:
      *r = 0;
      *g = x;
      *b = chroma;
   case 4:
      *r = x;
      *g = 0;
      *b = chroma;
   case 5:
      *r = chroma;
      *g = 0;
      *b = x;
   }

   int m = val - chroma;
   *r += m;
   *g += m;
   *b += m;

   //printf("{%d,%d,%d} => {%d,%d,%d}\n", hue, sat, val, *r, *g, *b);
}

Mesh* beat;
const float BEAT_SPEED = 80;
void initGeom()
{
   beat = GeometryCreator::CreateTorus(.85, 1, 12, 24);
}

float maxHighSoFar = 0.01;
// Returns float between 0 and 1
float getHighs()
{
   // get the FFT data
   float fft[256];
   BASS_ChannelGetData(chan, fft, BASS_DATA_FFT256);

   // get a sum representing high-frequency content
   float sum = 0;
   for (int i  = 64; i < 128; i++)
   {
      sum += fft[i];
   }
   if (sum > maxHighSoFar)
      maxHighSoFar = sum;

   return sum / maxHighSoFar;
}

// Gets how loud the sound is right now
int getLevel()
{
   DWORD level = BASS_ChannelGetLevel(chan);
   DWORD left  = LOWORD(level);
   DWORD right = HIWORD(level);
   int returnVal = (left + right) / 2;
   returnVal /= 128;
   return returnVal;
}

// returns band between 0 and 512
int getMaxBassBand()
{
   // get the FFT data
   float fft[16384];
   BASS_ChannelGetData(chan, fft, BASS_DATA_FFT16384);

   int max = 0;
   for (int i = 1; i < 512; i++)
   {
      if (fft[i] > fft[max])
         max = i;
   }
   return max;
}

void Timer(int param)
{
    time += 1;
    glutPostRedisplay();
}

void drawBeatDonut()
{
   ModelTrans.pushMatrix();

   ModelTrans.translate(vec3(0,0, (time - timeLastBeat) / BEAT_SPEED - 5));
   ModelTrans.scale((time - timeLastBeat) / BEAT_SPEED);
   SetModel();

   safe_glEnableVertexAttribArray(aPosition);
   glBindBuffer(GL_ARRAY_BUFFER, beat->PositionHandle);
   safe_glVertexAttribPointer(aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);

   safe_glEnableVertexAttribArray(aNormal);
   glBindBuffer(GL_ARRAY_BUFFER, beat->NormalHandle);
   safe_glVertexAttribPointer(aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);

   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, beat->IndexHandle);

   glDrawElements(GL_TRIANGLES, beat->IndexBufferLength, GL_UNSIGNED_SHORT, 0);

   ModelTrans.popMatrix();
}

void Draw()
{
   // Set background
   float bgBrightness = getHighs();
   glClearColor(bgBrightness, bgBrightness, bgBrightness, 0);

   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   glUseProgram(ShadeProg);

   SetProjectionMatrix();
   SetView();
   
   ModelTrans.loadIdentity();
   
   // Set donut color, default dark red
   int donutr = 128, donutg = 16, donutb = 255;
   // Map bands 15-40 to hues 245-375
   int maxBand = getMaxBassBand();
   //printf ("max band = %d\n", maxBand);
   if (15 <= maxBand && maxBand <= 50)
   {
      int bassHue = (int)((maxBand - 15) * (130.0/35.0) + 245);
      getRGB(bassHue % 360, 255, getLevel(), &donutr, &donutg, &donutb);
   }
   glUniform3f(uColor, donutr / 256.0, donutg / 256.0, donutb / 256.0);

   // Draw donut field
   for (int y = 5; y >= -5; y -= 1)
      for (int x = -10; x <= 10; x += 2)
      {
         ModelTrans.pushMatrix();
         ModelTrans.translate(vec3(x, y, 0));
         if (y % 2 == 0)
            ModelTrans.translate(vec3(1,0,0));
         drawBeatDonut();
         ModelTrans.popMatrix();
      }

   safe_glDisableVertexAttribArray(aPosition);
   safe_glDisableVertexAttribArray(aNormal);

   glUseProgram(0);
   glutSwapBuffers();
   printOpenGLError();

   glutTimerFunc(1, Timer, 1);
}

void Reshape(int width, int height)
{
    WindowWidth = width;
    WindowHeight = height;
    glViewport(0, 0, width, height);
}

void Keyboard(unsigned char key, int x, int y)
{
   QWORD pos;

   switch (key)
   {
   case 'q': case 'Q' :
      exit( EXIT_SUCCESS );
      break;
   case 'j':
      pos = BASS_ChannelGetPosition(chan, BASS_POS_BYTE);
      pos -= 48100 * 15; // Seek back a few seconds in song
      BASS_ChannelSetPosition(chan, pos, BASS_POS_BYTE);
      break;
   case 'k':
      pos = BASS_ChannelGetPosition(chan, BASS_POS_BYTE);
      pos += 48100 * 15; // Seek forward a few seconds in song
      BASS_ChannelSetPosition(chan, pos, BASS_POS_BYTE);
      break;
   case 't':
      // Trigger Beat
      timeLastBeat = time;
      break;
   }
}

#define DETECTION_LATENCY 20
void CALLBACK beatProcGetBPM(DWORD handle, double beatpos, void *user)
{
   timeLastBeat = time - DETECTION_LATENCY;
}

// From BASS example library
bool playFile()
{
	char file[MAX_PATH] = "";
	OPENFILENAME ofn = {0};
	ofn.lStructSize = sizeof(ofn);
	ofn.hwndOwner = win;
	ofn.nMaxFile = MAX_PATH;
	ofn.lpstrFile = file;
	ofn.Flags = OFN_FILEMUSTEXIST|OFN_HIDEREADONLY|OFN_EXPLORER;
	ofn.lpstrTitle = "Select a file to play";
	ofn.lpstrFilter = "playable files\0*.mo3;*.xm;*.mod;*.s3m;*.it;*.mtm;*.umx;*.mp3;*.mp2;*.mp1;*.ogg;*.wav;*.aif\0All files\0*.*\0\0";
	if (!GetOpenFileName(&ofn)) return FALSE;

	if (!(chan=BASS_StreamCreateFile(FALSE,file,0,0,BASS_SAMPLE_LOOP))
		&& !(chan=BASS_MusicLoad(FALSE,file,0,0,BASS_MUSIC_RAMP|BASS_SAMPLE_LOOP,1)))
   {
		printf("ERROR:Can't play file\n");
		return FALSE; // Can't load the file
	}

	BASS_ChannelPlay(chan, FALSE);

	return TRUE;
}

void initBASS()
{
	// setup output - default device, 44100hz, stereo, 16 bits
	if (!BASS_Init(-1,44100,0,win,NULL)) {
		printf("Can't initialize device\n");
		exit(1);
   }
   // Start a file playing
   if (!playFile()) {
	   BASS_Free();
	   exit(1);
   }

   // Set up beat callback
   BASS_FX_BPM_BeatCallbackSet(chan, &beatProcGetBPM, 0);
}

int main(int argc, char *argv[])
{
   // Glut Setup
   glutInit(& argc, argv);
   glutInitWindowPosition(100, 100);
   glutInitWindowSize(WindowWidth, WindowHeight);
   glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
   glutCreateWindow("Mind's Eye");
   glutDisplayFunc(Draw);
   glutReshapeFunc(Reshape);
   glutKeyboardFunc(Keyboard);
   glutTimerFunc(1, Timer, 1);

   // GLEW Setup (Windows only)
   #ifdef _WIN32
   GLenum err = glewInit();
   if (GLEW_OK != err)
   {
      std::cerr << "Error initializing glew! " << glewGetErrorString(err) << std::endl;
      return 1;
   }
   #endif

   // OpenGL Setup
   Initialize();
   initGeom();
   getGLversion();

   // Shader Setup
   if (! InstallShader("Diffuse.vert", "Diffuse.frag"))
   {
      printf("Error installing shader!\n");
      return 1;
   }

   initBASS();

   glutMainLoop();

   BASS_Free();

   return 0;
}
